@@ -1,24 +1,16 @@
- # ====================================================================
- # Copyright (c) 2007-2007 Open Source Applications Foundation.
- #
- # Permission is hereby granted, free of charge, to any person obtaining a
- # copy of this software and associated documentation files (the "Software"),
- # to deal in the Software without restriction, including without limitation
- # the rights to use, copy, modify, merge, publish, distribute, sublicense,
- # and/or sell copies of the Software, and to permit persons to whom the
- # Software is furnished to do so, subject to the following conditions: 
- #
- # The above copyright notice and this permission notice shall be included
- # in all copies or substantial portions of the Software. 
- #
- # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
- # OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
- # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- # FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- # DEALINGS IN THE SOFTWARE.
- # ====================================================================
+#   Copyright (c) 2007-2008 Open Source Applications Foundation
+#
+#   Licensed under the Apache License, Version 2.0 (the "License");
+#   you may not use this file except in compliance with the License.
+#   You may obtain a copy of the License at
+#
+#       http://www.apache.org/licenses/LICENSE-2.0
+#
+#   Unless required by applicable law or agreed to in writing, software
+#   distributed under the License is distributed on an "AS IS" BASIS,
+#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+#   See the License for the specific language governing permissions and
+#   limitations under the License.
 
 import os, sys, zipfile, _jcc
 
@@ -79,7 +71,8 @@ PRIMITIVES = { 'boolean': 'Z',
                'short': 'S',
                'void': 'V' }
 
-RESERVED = set(['delete', 'and', 'or', 'not', 'xor'])
+RESERVED = set(['delete', 'and', 'or', 'not', 'xor', 'union', 'NULL',
+                'register', 'const'])
 
 
 def typename(cls, current, const):
@@ -127,7 +120,7 @@ def line(out, indent=0, string='', *args):
 
     out.write(INDENT * indent)
     out.write(string % args)
-    out.write(os.linesep)
+    out.write('\n')
 
 
 def known(cls, typeset, declares, packages, excludes):
@@ -244,6 +237,7 @@ def jcc(args):
     jars = []
     classpath = []
     moduleName = None
+    modules = []
     build = False
     install = False
     recompile = False
@@ -257,6 +251,11 @@ def jcc(args):
     wrapperFiles = 1
     prefix = None
     root = None
+    install_dir = None
+    use_distutils = False
+    shared = False
+    dist = False
+    compiler = None
 
     i = 1
     while i < len(args):
@@ -267,6 +266,10 @@ def jcc(args):
                 classpath.append(args[i])
                 classNames.update(expandjar(args[i]))
                 jars.append(args[i])
+            elif arg == '--include':
+                i += 1
+                classpath.append(args[i])
+                jars.append(args[i])
             elif arg == '--package':
                 i += 1
                 packages.add(args[i])
@@ -277,6 +280,9 @@ def jcc(args):
                 from python import python, module
                 i += 1
                 moduleName = args[i]
+            elif arg == '--module':
+                i += 1
+                modules.append(args[i])
             elif arg == '--build':
                 from python import compile
                 build = True
@@ -286,7 +292,6 @@ def jcc(args):
             elif arg == '--compile':
                 from python import compile
                 recompile = True
-                install = True
             elif arg == '--output':
                 i += 1
                 output = args[i]
@@ -315,21 +320,40 @@ def jcc(args):
             elif arg == '--root':
                 i += 1
                 root = args[i]
+            elif arg == '--install-dir':
+                i += 1
+                install_dir = args[i]
+            elif arg == '--use-distutils':
+                use_distutils = True
+            elif arg == '--shared':
+                shared = True
+            elif arg == '--bdist':
+                from python import compile
+                dist = True
+            elif arg == '--compiler':
+                i += 1
+                compiler = args[i]
+            else:
+                raise ValueError, "Invalid argument: %s" %(arg)
         else:
             classNames.add(arg)
         i += 1
 
-    env = initVM(os.pathsep.join(classpath) or None)
+    env = initVM(os.pathsep.join(classpath) or None,
+                 vmargs='-Djava.awt.headless=true')
+
     typeset = set()
-    excludes = set([env.findClass(className.replace('.', '/'))
+    excludes = set([findClass(className.replace('.', '/'))
                     for className in excludes])
 
-    if recompile or not build and install:
+    if recompile or not build and (install or dist):
         compile(env, os.path.dirname(args[0]), output, moduleName,
-                install, debug, jars, version, prefix, root)
+                install, dist, debug, jars, version,
+                prefix, root, install_dir, use_distutils,
+                shared, compiler, modules)
     else:
         for className in classNames:
-            cls = env.findClass(className.replace('.', '/'))
+            cls = findClass(className.replace('.', '/'))
             if cls is None:
                 raise ValueError, className
             if cls in excludes:
@@ -340,11 +364,19 @@ def jcc(args):
                 while cls and cls not in typeset:
                     typeset.add(cls)
                     cls = cls.getSuperclass()
-        typeset.add(env.findClass('java/lang/Class'))
-        typeset.add(env.findClass('java/lang/RuntimeException'))
+        typeset.add(findClass('java/lang/Class'))
+        typeset.add(findClass('java/lang/String'))
+        typeset.add(findClass('java/lang/Throwable'))
+        typeset.add(findClass('java/lang/Exception'))
+        typeset.add(findClass('java/lang/RuntimeException'))
         if moduleName:
-            typeset.add(env.findClass('java/util/Iterator'))
-            typeset.add(env.findClass('java/util/Enumeration'))
+            typeset.add(findClass('java/lang/Number'))
+            typeset.add(findClass('java/lang/Boolean'))
+            typeset.add(findClass('java/lang/Integer'))
+            typeset.add(findClass('java/lang/Long'))
+            typeset.add(findClass('java/lang/Double'))
+            typeset.add(findClass('java/util/Iterator'))
+            typeset.add(findClass('java/util/Enumeration'))
 
         if moduleName:
             cppdir = os.path.join(output, '_%s' %(moduleName))
@@ -422,11 +454,13 @@ def jcc(args):
 
         if moduleName:
             out = file(os.path.join(cppdir, moduleName) + '.cpp', 'w')
-            module(out, allInOne, done, cppdir, moduleName)
+            module(out, allInOne, done, cppdir, moduleName, shared)
             out.close()
-            if build or install:
+            if build or install or dist:
                 compile(env, os.path.dirname(args[0]), output, moduleName,
-                        install, debug, jars, version, prefix, root)
+                        install, dist, debug, jars, version,
+                        prefix, root, install_dir, use_distutils,
+                        shared, compiler, modules)
 
 
 def header(env, out, cls, typeset, packages, excludes):
@@ -446,7 +480,7 @@ def header(env, out, cls, typeset, packages, excludes):
         if interfaces:
             superCls = interfaces.pop(0)
         else:
-            superCls = env.findClass('java/lang/Object')
+            superCls = findClass('java/lang/Object')
         superClsName = superCls.getName()
     elif superCls:
         superClsName = superCls.getName()
@@ -622,6 +656,7 @@ def header(env, out, cls, typeset, packages, excludes):
         for field in instanceFields:
             fieldType = field.getType()
             fieldName = field.getName()
+            modifiers = field.getModifiers()
             line(out, indent, '%s _get_%s() const;',
                  typename(fieldType, cls, False), fieldName)
             if not Modifier.isFinal(modifiers):
