@@ -1,46 +1,29 @@
- # ====================================================================
- # Copyright (c) 2007-2007 Open Source Applications Foundation.
- #
- # Permission is hereby granted, free of charge, to any person obtaining a
- # copy of this software and associated documentation files (the "Software"),
- # to deal in the Software without restriction, including without limitation
- # the rights to use, copy, modify, merge, publish, distribute, sublicense,
- # and/or sell copies of the Software, and to permit persons to whom the
- # Software is furnished to do so, subject to the following conditions: 
- #
- # The above copyright notice and this permission notice shall be included
- # in all copies or substantial portions of the Software. 
- #
- # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
- # OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
- # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- # FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- # DEALINGS IN THE SOFTWARE.
- # ====================================================================
+#   Copyright (c) 2007-2008 Open Source Applications Foundation
+#
+#   Licensed under the Apache License, Version 2.0 (the "License");
+#   you may not use this file except in compliance with the License.
+#   You may obtain a copy of the License at
+#
+#       http://www.apache.org/licenses/LICENSE-2.0
+#
+#   Unless required by applicable law or agreed to in writing, software
+#   distributed under the License is distributed on an "AS IS" BASIS,
+#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+#   See the License for the specific language governing permissions and
+#   limitations under the License.
 
-import os, sys, shutil
+import os, sys, shutil, _jcc, py_compile
 
 from cpp import PRIMITIVES, RESERVED, INDENT, HALF_INDENT
 from cpp import typename, line, signature, find_method, split_pkg, sort
 from cpp import Modifier
-from config import INCLUDES, LFLAGS
+from _jcc import findClass
+from config import INCLUDES, CFLAGS, DEBUG_CFLAGS, LFLAGS, SHARED
 
 python_ver = '%d.%d.%d' %(sys.version_info[0:3])
 if python_ver < '2.4':
     from sets import Set as set
 
-try:
-    from setuptools import setup, Extension
-    with_setuptools = True
-except ImportError:
-    if python_ver < '2.4':
-        raise ImportError, 'setuptools is required when using Python 2.3'
-    else:
-        from distutils.core import setup, Extension
-        with_setuptools = False
-
 
 RESULTS = { 'boolean': 'Py_RETURN_BOOL(%s);',
             'byte': 'return PyString_FromStringAndSize((char *) &%s, 1);',
@@ -50,9 +33,7 @@ RESULTS = { 'boolean': 'Py_RETURN_BOOL(%s);',
             'int': 'return PyInt_FromLong((long) %s);',
             'long': 'return PyLong_FromLongLong((PY_LONG_LONG) %s);',
             'short': 'return PyInt_FromLong((long) %s);',
-            'java.lang.String': 'return j2p(%s);',
-            '[char': 'return %s.toSequence();',
-            '[byte': 'return %s.toSequence();' }
+            'java.lang.String': 'return j2p(%s);' }
 
 CALLARGS = { 'boolean': ('O', '(%s ? Py_True : Py_False)', False),
              'byte': ('O', 'PyString_FromStringAndSize((char *) &%s, 1)', True),
@@ -62,9 +43,7 @@ CALLARGS = { 'boolean': ('O', '(%s ? Py_True : Py_False)', False),
              'int': ('i', '(int) %s', False),
              'long': ('L', '(long long) %s', False),
              'short': ('i', '(int) %s', False),
-             'java.lang.String': ('O', 'env->fromJString((jstring) %s)', True),
-             '[char': ('O', 'JArray<jchar>(%s).toSequence()', True),
-             '[byte': ('O', 'JArray<jbyte>(%s).toSequence()', True) }
+             'java.lang.String': ('O', 'env->fromJString((jstring) %s)', True) }
 
 
 def parseArgs(params, current):
@@ -74,16 +53,20 @@ def parseArgs(params, current):
         while cls.isArray():
             array += '['
             cls = cls.getComponentType()
+        clsName = cls.getName()
         if cls.isPrimitive():
-            return array + PRIMITIVES[cls.getName()]
-        if cls.getName() == 'java.lang.String':
+            return array + PRIMITIVES[clsName]
+        if clsName == 'java.lang.String':
             return array + 's'
+        if clsName == 'java.lang.Object':
+            return array + 'o'
         return array + 'j'
 
     def checkarg(cls):
         while cls.isArray():
             cls = cls.getComponentType()
-        if (cls.isPrimitive() or cls.getName() == 'java.lang.String'):
+        if (cls.isPrimitive() or
+            cls.getName() in ('java.lang.String', 'java.lang.Object')):
             return ''
         return ', %s::getclass$()' %(typename(cls, current, False))
 
@@ -148,21 +131,21 @@ def fieldValue(cls, value, fieldType):
     if fieldType.isArray():
         fieldType = fieldType.getComponentType()
         if fieldType.isArray():
-            raise NotImplementedError, 'nested arrays'
+            result = 'JArray<jobject>(%s->this$).wrap(NULL)'
         elif fieldType.isPrimitive():
-            result = '%s->toSequence()'
+            result = '%s->wrap()'
         elif fieldType.getName() == 'java.lang.String':
-            result = '%s->toSequence(j2p)'
+            result = 'JArray<jstring>(%s->this$).wrap()'
         else:
             parts = rpartition(typename(fieldType, cls, False), '::')
-            result = "%%s->toSequence(%s%st_%s::wrapObject)" %(parts)
+            result = 'JArray<jobject>(%%s->this$).wrap(%s%st_%s::wrap_jobject)' %(parts)
 
     elif fieldType.getName() == 'java.lang.String':
         result = 'j2p(*%s)'
 
     elif not fieldType.isPrimitive():
         parts = rpartition(typename(fieldType, cls, False), '::')
-        result = '%s%st_%s::wrapObject(*%%s)' %(parts)
+        result = '%s%st_%s::wrap_Object(*%%s)' %(parts)
 
     else:
         return value
@@ -181,19 +164,19 @@ def returnValue(cls, returnType, value):
                 returnType = returnType.getComponentType()
                 depth += 1
             if depth > 1:
-                result = 'Py_RETURN_NONE; //todo: %s'
+                result = 'return JArray<jobject>(%s.this$).wrap(NULL);'
             elif returnType.isPrimitive():
-                result = 'return %s.toSequence();'
+                result = 'return %s.wrap();'
             elif returnType.getName() == 'java.lang.String':
-                result = 'return %s.toSequence(j2p);'
+                result = 'return JArray<jstring>(%s.this$).wrap();'
             else:
                 returnName = typename(returnType, cls, False)
                 parts = rpartition(returnName, '::')
-                result = 'return %%s.toSequence(%s%st_%s::wrapObject);' %(parts)
+                result = 'return JArray<jobject>(%%s.this$).wrap(%s%st_%s::wrap_jobject);' %(parts)
         else:
             returnName = typename(returnType, cls, False)
             parts = rpartition(returnName, '::')
-            result = 'return %s%st_%s::wrapObject(%%s);' %(parts)
+            result = 'return %s%st_%s::wrap_Object(%%s);' %(parts)
 
     return result %(value)
 
@@ -247,10 +230,18 @@ def call(out, indent, cls, inCase, method, names, cardinality, isExtension):
     if isExtension and name == 'clone' and Modifier.isNative(modifiers):
         line(out)
         line(out, indent, '%s object(result.this$);', typename(cls, cls, False))
-        line(out, indent, 'Py_INCREF((PyObject *) arg);')
-        line(out, indent, 'object.pythonExtension((jlong) (Py_intptr_t) (void *) arg);')
-
-    if returnName != 'void':
+        line(out, indent, 'if (PyObject_TypeCheck(arg, &FinalizerProxyType) &&')
+        line(out, indent, '    PyObject_TypeCheck(((t_fp *) arg)->object, self->ob_type))')
+        line(out, indent, '{')
+        line(out, indent + 1, 'PyObject *_arg = ((t_fp *) arg)->object;')
+        line(out, indent + 1, '((t_JObject *) _arg)->object = object;')
+        line(out, indent + 1, 'Py_INCREF(_arg);')
+        line(out, indent + 1, 'object.pythonExtension((jlong) (Py_intptr_t) (void *) _arg);')
+        line(out, indent + 1, 'Py_INCREF(arg);')
+        line(out, indent + 1, 'return arg;')
+        line(out, indent, '}')
+        line(out, indent, 'return PyErr_SetArgsError("%s", arg);' %(name))
+    elif returnName != 'void':
         line(out, indent, returnValue(cls, returnType, 'result'))
     else:
         line(out, indent, 'Py_RETURN_NONE;')
@@ -304,14 +295,14 @@ def extension(env, out, indent, cls, names, name, count, method):
          names[-1], names[-1], env.strhash('()J'))
     line(out, indent, 'PyObject *obj = (PyObject *) (Py_intptr_t) ptr;')
 
-    if name == 'pythonIncRef':
-        line(out, indent, 'PythonGIL gil;')
-        line(out, indent, 'Py_INCREF(obj);')
-        return
-
     if name == 'pythonDecRef':
-        line(out, indent, 'PythonGIL gil;')
-        line(out, indent, 'Py_DECREF(obj);')
+        line(out)
+        line(out, indent, 'if (obj != NULL)')
+        line(out, indent, '{')
+        line(out, indent + 1, 'jenv->CallVoidMethod(jobj, %s::mids$[%s::mid_pythonExtension_%s], (jlong) 0);',
+             names[-1], names[-1], env.strhash('(J)V'))
+        line(out, indent + 1, 'env->finalizeObject(jenv, obj);')
+        line(out, indent, '}')
         return
 
     returnType = method.getReturnType()
@@ -321,7 +312,7 @@ def extension(env, out, indent, cls, names, name, count, method):
              typename(returnType, cls, False),
              not returnType.isPrimitive() and '((jobject) NULL)' or '')
 
-    line(out, indent, 'PythonGIL gil;')
+    line(out, indent, 'PythonGIL gil(jenv);')
 
     sigs = []
     decrefs = []
@@ -334,20 +325,20 @@ def extension(env, out, indent, cls, names, name, count, method):
         elif param.isArray():
             param = param.getComponentType()
             if param.isPrimitive():
-                code = 'JArray<j%s>(%%s).toSequence()' %(param.getName())
+                code = 'JArray<j%s>(%%s).wrap()' %(param.getName())
             elif param.isArray():
-                raise NotImplementedError, 'nested arrays'
+                code = 'JArray<jobject>(%s).wrap(NULL)'
             elif param.getName() == 'java.lang.String':
-                code = 'JArray<java::lang::String>.toSequence(j2p)(%s)'
+                code = 'JArray<jstring>(%s).wrap()'
             else:
                 parts = rpartition(typename(param, cls, False), '::')
-                code = 'JArray<%s%s%s>.toSequence(%s%st_%s::wrapObject)(%%s)' %(parts*2)
+                code = 'JArray<jobject>(%%s).wrap(%s%st_%s::wrap_jobject)' %(parts)
             sig, decref = 'O', True
         elif param.getName() == 'java.lang.String':
             sig, code, decref = 'O', 'j2p(%%s))', True
         else:
             parts = rpartition(typename(param, cls, False), '::')
-            sig, code, decref = 'O', '%s%st_%s::wrapObject(%s%s%s(%%s))' %(parts*2), True
+            sig, code, decref = 'O', '%s%st_%s::wrap_Object(%s%s%s(%%s))' %(parts*2), True
         if sig == 'O':
             line(out, indent, 'PyObject *o%d = %s;', i, code %('a%d' %(i)))
             args.append('o%d' %(i))
@@ -416,14 +407,15 @@ def python(env, out_h, out, cls, superCls, names, superNames,
     line(out_h, indent, 'public:')
     line(out_h, indent + 1, 'PyObject_HEAD')
     line(out_h, indent + 1, '%s object;', names[-1])
-    line(out_h, indent + 1, 'static PyObject *wrapObject(const %s&);',
+    line(out_h, indent + 1, 'static PyObject *wrap_Object(const %s&);',
          names[-1])
+    line(out_h, indent + 1, 'static PyObject *wrap_jobject(const jobject&);')
     line(out_h, indent + 1, 'static void install(PyObject *module);')
     line(out_h, indent + 1, 'static void initialize(PyObject *module);')
     line(out_h, indent, '};')
 
-    iterator = env.findClass('java/util/Iterator')
-    enumeration = env.findClass('java/util/Enumeration')
+    iterator = findClass('java/util/Iterator')
+    enumeration = findClass('java/util/Enumeration')
 
     while indent:
         indent -= 1
@@ -514,9 +506,12 @@ def python(env, out_h, out, cls, superCls, names, superNames,
     allMethods = allMethods.items()
     sort(allMethods, key=lambda x: x[0])
 
-    hasIterator = False
+    iteratorMethod = None
+    iteratorExt = False
     nextMethod = None
+    nextExt = False
     nextElementMethod = None
+    nextElementExt = False
 
     mappingMethod = None
     if mapping:
@@ -533,9 +528,10 @@ def python(env, out_h, out, cls, superCls, names, superNames,
         sort(methods, key=lambda x: len(x.getParameterTypes()))
         method = methods[0]
         modifiers = method.getModifiers()
-        if name == 'iterator' and not hasIterator:
-            hasIterator = (not method.getParameterTypes() and
-                           iterator.isAssignableFrom(method.getReturnType()))
+        if name == 'iterator' and iteratorMethod is None:
+            if (not method.getParameterTypes() and
+                iterator.isAssignableFrom(method.getReturnType())):
+                iteratorMethod = method
         elif name == 'next' and nextMethod is None:
             if (not method.getParameterTypes() and
                 not method.getReturnType().isPrimitive()):
@@ -563,6 +559,27 @@ def python(env, out_h, out, cls, superCls, names, superNames,
             line(out, indent, 'static PyObject *t_%s_%s(t_%s *self%s);',
                  names[-1], name, names[-1], args)
 
+    for name, methods in extMethods:
+        args, x, cardinality = methodargs(methods, superMethods)
+        sort(methods, key=lambda x: len(x.getParameterTypes()))
+        method = methods[0]
+        modifiers = method.getModifiers()
+        if name == 'iterator' and iteratorMethod is None:
+            if (not method.getParameterTypes() and
+                iterator.isAssignableFrom(method.getReturnType())):
+                iteratorMethod = method
+                iteratorExt = True
+        elif name == 'next' and nextMethod is None:
+            if (not method.getParameterTypes() and
+                not method.getReturnType().isPrimitive()):
+                nextMethod = method
+                nextExt = True
+        elif name == 'nextElement' and nextElementMethod is None:
+            if (not method.getParameterTypes() and
+                not method.getReturnType().isPrimitive()):
+                nextElementMethod = method
+                nextElementExt = True
+
     if isExtension:
         count = 0
         for name, methods in extMethods:
@@ -572,6 +589,7 @@ def python(env, out_h, out, cls, superCls, names, superNames,
                      jniname(method.getReturnType()), names[-1], name, count,
                      jniargs(method.getParameterTypes()))
                 count += 1
+        line(out, indent, 'static PyObject *t_%s_get__self(t_%s *self, void *data);', names[-1], names[-1])
 
     if instanceFields:
         for field in instanceFields:
@@ -581,7 +599,7 @@ def python(env, out_h, out, cls, superCls, names, superNames,
                      names[-1], fieldName, names[-1])
                 if not Modifier.isFinal(field.getModifiers()):
                     line(out, indent, 'static int t_%s_set__%s(t_%s *self, PyObject *arg, void *data);',
-                     names[-1], field.getName(), names[-1])
+                         names[-1], field.getName(), names[-1])
         line(out)
 
     for fieldName, methods in propMethods:
@@ -599,7 +617,7 @@ def python(env, out_h, out, cls, superCls, names, superNames,
                 line(out, indent, 'static int t_%s_set__%s(t_%s *self, PyObject *arg, void *data);',
                      names[-1], fieldName, names[-1])
 
-    if instanceFields or propMethods:
+    if instanceFields or propMethods or isExtension:
         line(out, indent, 'static PyGetSetDef t_%s__fields_[] = {', names[-1])
         for field in instanceFields:
             fieldName = field.getName()
@@ -628,6 +646,8 @@ def python(env, out_h, out, cls, superCls, names, superNames,
                     op = 'SET'
             line(out, indent + 1, 'DECLARE_%s_FIELD(t_%s, %s),',
                  op, names[-1], fieldName)
+        if isExtension:
+            line(out, indent + 1, 'DECLARE_GET_FIELD(t_%s, self),', names[-1])
         line(out, indent + 1, '{ NULL, NULL, NULL, NULL, NULL }')
         line(out, indent, '};')
 
@@ -662,29 +682,41 @@ def python(env, out_h, out, cls, superCls, names, superNames,
     line(out, indent + 1, '{ NULL, NULL, 0, NULL }')
     line(out, indent, '};')
 
-    if instanceFields or propMethods:
+    if instanceFields or propMethods or isExtension:
         tp_getset = 't_%s__fields_' %(names[-1])
     else:
         tp_getset = '0'
 
-    if hasIterator:
-        tp_iter = 'get_iterator<t_%s>' %(names[-1])
+    if iteratorMethod:
+        if iteratorExt:
+            tp_iter = 'get_extension_iterator'
+        else:
+            tp_iter = 'get_iterator<t_%s>' %(names[-1])
         tp_iternext = '0'
-    elif iterator.isAssignableFrom(cls):
+    elif nextMethod and iterator.isAssignableFrom(cls):
         tp_iter = 'PyObject_SelfIter'
         returnName = typename(nextMethod.getReturnType(), cls, False)
         ns, sep, n = rpartition(returnName, '::')
-        tp_iternext = '(get_iterator_next<%s%st_%s,%s>)' %(ns, sep, n, returnName)
-    elif enumeration.isAssignableFrom(cls):
+        if nextExt:
+            tp_iternext = 'get_extension_next'
+        else:
+            tp_iternext = '(get_iterator_next<%s%st_%s,%s>)' %(ns, sep, n, returnName)
+    elif nextElementMethod and enumeration.isAssignableFrom(cls):
         tp_iter = 'PyObject_SelfIter'
         returnName = typename(nextElementMethod.getReturnType(), cls, False)
         ns, sep, n = rpartition(returnName, '::')
-        tp_iternext = '(get_enumeration_nextElement<%s%st_%s,%s>)' %(ns, sep, n, returnName)
+        if nextElementExt:
+            tp_iternext = 'get_extension_nextElement'
+        else:
+            tp_iternext = '(get_enumeration_nextElement<%s%st_%s,%s>)' %(ns, sep, n, returnName)
     elif nextMethod:
         tp_iter = 'PyObject_SelfIter'
         returnName = typename(nextMethod.getReturnType(), cls, False)
         ns, sep, n = rpartition(returnName, '::')
-        tp_iternext = '(get_next<t_%s,%s%st_%s,%s>)' %(names[-1], ns, sep, n, returnName)
+        if nextExt:
+            tp_iternext = 'get_extension_next'
+        else:
+            tp_iternext = '(get_next<t_%s,%s%st_%s,%s>)' %(names[-1], ns, sep, n, returnName)
     else:
         tp_iter = '0'
         tp_iternext = '0'
@@ -752,7 +784,10 @@ def python(env, out_h, out, cls, superCls, names, superNames,
     line(out)
     line(out, indent, 'void t_%s::install(PyObject *module)', names[-1])
     line(out, indent, '{')
-    line(out, indent + 1, 'INSTALL_TYPE(%s, module);', names[-1])
+    if isExtension:
+        line(out, indent + 1, 'INSTALL_EXTENSION(%s, module);', names[-1])
+    else:
+        line(out, indent + 1, 'INSTALL_TYPE(%s, module);', names[-1])
     for inner in cls.getDeclaredClasses():
         if inner in typeset:
             if Modifier.isStatic(inner.getModifiers()):
@@ -766,6 +801,8 @@ def python(env, out_h, out, cls, superCls, names, superNames,
     line(out, indent, '{')
     line(out, indent + 1, 'PyDict_SetItemString(%sType.tp_dict, "class_", make_descriptor(%s::initializeClass));',
          names[-1], names[-1])
+    line(out, indent + 1, 'PyDict_SetItemString(%sType.tp_dict, "wrapfn_", make_descriptor(t_%s::wrap_jobject));',
+         names[-1], names[-1])
 
     if isExtension:
         line(out, indent + 1, 'jclass cls = %s::initializeClass();', names[-1]);
@@ -799,14 +836,14 @@ def python(env, out_h, out, cls, superCls, names, superNames,
         line(out)
         line(out, indent, 'static PyObject *t_%s_cast_(PyTypeObject *type, PyObject *arg)', names[-1])
         line(out, indent, '{')
-        line(out, indent + 1, 'if (castCheck(arg, %s::initializeClass(), 1) < 0)', names[-1])
+        line(out, indent + 1, 'if (!(arg = castCheck(arg, %s::initializeClass(), 1)))', names[-1])
         line(out, indent + 2, 'return NULL;')
-        line(out, indent + 1, 'return t_%s::wrapObject(%s(((t_%s *) arg)->object.this$));', names[-1], names[-1], names[-1])
+        line(out, indent + 1, 'return t_%s::wrap_Object(%s(((t_%s *) arg)->object.this$));', names[-1], names[-1], names[-1])
         line(out, indent, '}')
 
         line(out, indent, 'static PyObject *t_%s_instance_(PyTypeObject *type, PyObject *arg)', names[-1])
         line(out, indent, '{')
-        line(out, indent + 1, 'if (castCheck(arg, %s::initializeClass(), 0) < 0)', names[-1])
+        line(out, indent + 1, 'if (!castCheck(arg, %s::initializeClass(), 0))', names[-1])
         line(out, indent + 2, 'Py_RETURN_FALSE;')
         line(out, indent + 1, 'Py_RETURN_TRUE;')
         line(out, indent, '}')
@@ -914,6 +951,24 @@ def python(env, out_h, out, cls, superCls, names, superNames,
                 line(out, indent, '{')
                 extension(env, out, indent + 1, cls, names, name, count, method)
                 line(out, indent, '}')
+        line(out)
+        line(out, indent, 'static PyObject *t_%s_get__self(t_%s *self, void *data)',
+             names[-1], names[-1])
+        line(out, indent, '{')
+        indent += 1
+        line(out, indent, 'jlong ptr;')
+        line(out, indent, 'OBJ_CALL(ptr = self->object.pythonExtension());')
+        line(out, indent, 'PyObject *obj = (PyObject *) (Py_intptr_t) ptr;')
+        line(out)
+        line(out, indent, 'if (obj != NULL)')
+        line(out, indent, '{')
+        line(out, indent + 1, 'Py_INCREF(obj);')
+        line(out, indent + 1, 'return obj;')
+        line(out, indent, '}')
+        line(out, indent, 'else')
+        line(out, indent + 1, 'Py_RETURN_NONE;')
+        indent -= 1
+        line(out, indent, '}')
 
     if instanceFields:
         for field in instanceFields:
@@ -1059,9 +1114,15 @@ def package(out, allInOne, cppdir, namespace, names):
     if not allInOne:
         out = file(os.path.join(os.path.join(cppdir, *names),
                                 '__init__.cpp'), 'w')
-    line(out, 0, '#include <jni.h>')
-    line(out, 0, '#include <Python.h>')
-    line(out, 0, '#include "macros.h"')
+
+    if allInOne and not names or not allInOne:
+        line(out, 0, '#include <jni.h>')
+        line(out, 0, '#include <Python.h>')
+        line(out, 0, '#include "macros.h"')
+
+    if not names:
+        line(out)
+        line(out, 0, 'PyObject *initVM(PyObject *module, PyObject *args, PyObject *kwds);')
 
     packages = []
     types = []
@@ -1070,7 +1131,10 @@ def package(out, allInOne, cppdir, namespace, names):
     sort(namespaces, key=lambda x: x[0])
     for name, entries in namespaces:
         if entries is True:
-            line(out, 0, '#include "%s/%s.h"', '/'.join(names), name)
+            if names:
+                line(out, 0, '#include "%s/%s.h"', '/'.join(names), name)
+            else:
+                line(out, 0, '#include "%s.h"', name)
             types.append(name)
         else:
             packages.append((name, entries))
@@ -1099,12 +1163,24 @@ def package(out, allInOne, cppdir, namespace, names):
     line(out, indent, '}')
 
     line(out)
-    line(out, indent, 'void __initialize__(PyObject *module)')
-    line(out, indent, '{')
+    if not names:
+        line(out, indent, 'PyObject *__initialize__(PyObject *module, PyObject *args, PyObject *kwds)')
+        line(out, indent, '{')
+        line(out, indent + 1, 'PyObject *env = initVM(module, args, kwds);')
+        line(out)
+        line(out, indent + 1, 'if (env == NULL)')
+        line(out, indent + 2, 'return NULL;')
+        line(out)
+    else:
+        line(out, indent, 'void __initialize__(PyObject *module)')
+        line(out, indent, '{')
     for name in types:
         line(out, indent + 1, 't_%s::initialize(module);', name)
     for name, entries in packages:
         line(out, indent + 1, '%s::__initialize__(module);', name)
+    if not names:
+        line(out)
+        line(out, indent + 1, 'return env;')
     line(out, indent, '}')
 
     while indent:
@@ -1113,16 +1189,19 @@ def package(out, allInOne, cppdir, namespace, names):
 
     if not allInOne:
         out.close()
+    else:
+        line(out)
 
     for name, entries in packages:
         package(out, allInOne, cppdir, entries, names + (name,))
 
 
-def module(out, allInOne, classes, cppdir, moduleName):
+def module(out, allInOne, classes, cppdir, moduleName, shared):
 
     extname = '_%s' %(moduleName)
     line(out, 0, '#include <Python.h>')
     line(out, 0, '#include "macros.h"')
+    line(out, 0, '#include "jccfuncs.h"')
 
     if allInOne:
         out_init = file(os.path.join(cppdir, '__init__.cpp'), 'w')
@@ -1141,27 +1220,56 @@ def module(out, allInOne, classes, cppdir, moduleName):
 
     line(out)
     line(out, 0, 'void __install__(PyObject *module);')
-    line(out, 0, 'extern PyTypeObject JObjectType, ConstVariableDescriptorType, JCCEnvType;')
-    line(out, 0, 'extern PyMethodDef module_funcs[];')
+    line(out, 0, 'extern PyTypeObject JObjectType, ConstVariableDescriptorType, FinalizerClassType, FinalizerProxyType;')
+    line(out, 0, 'extern void _install_jarray(PyObject *);')
+    line(out)
+    line(out, 0, '#if defined(_jcc_shared) && (defined(_MSC_VER) || defined(__WIN32))')
+    line(out, 0, '#define _DLL_IMPORT __declspec(dllimport)')
+    line(out, 0, '#else')
+    line(out, 0, '#define _DLL_IMPORT')
+    line(out, 0, '#endif')
+    line(out, 0, '_DLL_IMPORT extern PyTypeObject JCCEnvType;')
+
     line(out)
     line(out, 0, 'extern "C" {')
 
     line(out)
     line(out, 1, 'void init%s(void)', extname)
     line(out, 1, '{')
-    line(out, 2, 'PyObject *module = Py_InitModule3("%s", module_funcs, "");',
+    line(out, 2, 'PyObject *module = Py_InitModule3("%s", jcc_funcs, "");',
          extname);
     line(out)
+    line(out, 2, 'PyEval_InitThreads();')
+    line(out)
     line(out, 2, 'INSTALL_TYPE(JObject, module);')
-    line(out, 2, 'INSTALL_TYPE(ConstVariableDescriptor, module);')
     line(out, 2, 'INSTALL_TYPE(JCCEnv, module);')
+    line(out, 2, 'INSTALL_TYPE(ConstVariableDescriptor, module);')
+    line(out, 2, 'INSTALL_TYPE(FinalizerClass, module);')
+    line(out, 2, 'INSTALL_TYPE(FinalizerProxy, module);')
+    line(out, 2, '_install_jarray(module);')
     line(out, 2, '__install__(module);')
     line(out, 1, '}')
     line(out, 0, '}')
 
 
-def compile(env, jccPath, output, moduleName, install, debug, jars, version,
-            prefix, root):
+def compile(env, jccPath, output, moduleName, install, dist, debug, jars,
+            version, prefix, root, install_dir, use_distutils,
+            shared, compiler, modules):
+
+    try:
+        if use_distutils:
+            raise ImportError
+        from setuptools import setup, Extension
+        with_setuptools = True
+        if shared and not SHARED:
+            raise NotImplementedError, "JCC was not built with --shared mode support, see JCC's INSTALL file for more information"
+    except ImportError:
+        if python_ver < '2.4':
+            raise ImportError, 'setuptools is required when using Python 2.3'
+        if shared:
+            raise ImportError, 'setuptools is required when using --shared'
+        from distutils.core import setup, Extension
+        with_setuptools = False
 
     extname = '_%s' %(moduleName)
 
@@ -1171,12 +1279,32 @@ def compile(env, jccPath, output, moduleName, install, debug, jars, version,
 
     out = file(os.path.join(modulePath, '__init__.py'), 'w')
     line(out)
-    line(out, 0, 'import os, %s', extname)
+    if shared:
+        line(out, 0, "import os, sys")
+        line(out)
+        line(out, 0, "if sys.platform == 'win32':")
+        line(out, 1, "import jcc, %s", extname)
+        line(out, 0, "else:")
+        line(out, 1, "import %s", extname)
+    else:
+        line(out, 0, 'import os, %s', extname)
     line(out)
-    line(out, 0, '__path__ = os.path.abspath(os.path.dirname(__file__))')
+    line(out, 0, '__dir__ = os.path.abspath(os.path.dirname(__file__))')
 
+    package_data = []
     for jar in jars:
         shutil.copy2(jar, modulePath)
+        package_data.append(os.path.basename(jar))
+
+    if modules:
+        for module in modules:
+            pfile = module.split('.')[0] + '.py'
+            shutil.copy2(pfile, modulePath)
+            pfile = os.path.basename(pfile)
+            cfile = pfile + (__debug__ and 'c' or 'o')
+            py_compile.compile(os.path.join(modulePath, pfile),
+                               os.path.join(modulePath, cfile),
+                               doraise=True)
 
     line(out)
     line(out, 0, 'class JavaError(Exception):')
@@ -1190,12 +1318,10 @@ def compile(env, jccPath, output, moduleName, install, debug, jars, version,
 
     if version:
         line(out)
-        line(out, 0, 'VERSION="%s"', version)
+        line(out, 0, 'VERSION = "%s"', version)
         
-    line(out, 0, 'CLASSPATH="%s" %%((__path__,)*%d)',
-         os.pathsep.join([os.path.join("%s", os.path.basename(jar))
-                          for jar in jars]),
-         len(jars))
+    line(out, 0, 'CLASSPATH = [%s]' %(', '.join(['os.path.join(__dir__, "%s")' %(os.path.basename(jar)) for jar in jars])))
+    line(out, 0, 'CLASSPATH = os.pathsep.join(CLASSPATH)')
 
     line(out)
     line(out, 0, 'from %s import *', extname)
@@ -1204,42 +1330,84 @@ def compile(env, jccPath, output, moduleName, install, debug, jars, version,
     includes = [os.path.join(output, extname),
                 os.path.join(jccPath, 'sources')]
 
-    sources = ['jcc.cpp', 'JCCEnv.cpp', 'JObject.cpp', 'functions.cpp']
-    sources = [os.path.join(jccPath, 'sources', source) for source in sources]
+    sources = ['JObject.cpp', 'JArray.cpp', 'functions.cpp', 'types.cpp']
+    if not shared:
+        sources.append('jcc.cpp')
+        sources.append('JCCEnv.cpp')
+    for source in sources:
+	shutil.copy2(os.path.join(jccPath, 'sources', source),
+                     os.path.join(output, extname))
+
+    sources = []
     for path, dirs, names in os.walk(os.path.join(output, extname)):
         for name in names:
             if name.endswith('.cpp'):
                 sources.append(os.path.join(path, name))
 
-    script_args = ['build_ext', '--define=PYTHON']
+    script_args = ['build_ext']
 
-    link_args = LFLAGS
     includes[0:0] = INCLUDES
+    compile_args = CFLAGS
+    link_args = LFLAGS
+    defines=['PYTHON']
+
+    if compiler:
+        script_args.append('--compiler=%s' %(compiler))
+
+    if shared:
+        defines.append('_jcc_shared')
+    script_args.append('--define=%s' %(','.join(defines)))
 
     if debug:
         script_args.append('--debug')
-    elif sys.platform != 'win32':
+        compile_args += DEBUG_CFLAGS
+    elif sys.platform == 'win32':
+        pass
+    elif sys.platform == 'sunos5':
+        link_args.append('-Wl,-s')
+    else:
         link_args.append('-Wl,-S')
 
     if install:
         script_args.append('install')
-
     if prefix:
         script_args.append('--prefix=%s' % prefix)
-
     if root:
         script_args.append('--root=%s' % root)
+    if install_dir:
+        script_args.append('--install-lib=%s' % install_dir)
 
-    extensions = [Extension('.'.join([moduleName, extname]),
-                            extra_link_args=link_args,
-                            include_dirs=includes,
-                            sources=sources)]
+    if dist:
+        if with_setuptools:
+            script_args.append('bdist_egg')
+        else:
+            script_args.append('bdist')
+
+    args = {
+        'extra_compile_args': compile_args,
+        'extra_link_args': link_args,
+        'include_dirs': includes,
+        'sources': sources
+    }
+    if shared:
+        shlibdir = os.path.dirname(os.path.dirname(_jcc.__file__))
+        if sys.platform in ('darwin', 'linux2'): # distutils no good with -R
+            args['extra_link_args'] += ['-Wl,-rpath', shlibdir]
+            args['library_dirs'] = [shlibdir]
+            args['libraries'] = ['jcc']
+        elif sys.platform == 'win32':
+            jcclib = 'jcc%s.lib' %(debug and '_d' or '')
+            args['extra_link_args'] += [os.path.join(shlibdir, 'jcc', jcclib)]
+        else:
+            raise NotImplementedError, "shared mode on %s" %(sys.platform)
+
+    extensions = [Extension('.'.join([moduleName, extname]), **args)]
 
     args = {
         'name': moduleName,
         'packages': [moduleName],
         'package_dir': {moduleName: modulePath},
-        'package_data': {moduleName: [os.path.basename(jar) for jar in jars]},
+        'package_data': {moduleName: package_data},
         'version': version,
         'ext_modules': extensions,
         'script_args': script_args
