@@ -1,45 +1,34 @@
-/* ====================================================================
- * Copyright (c) 2007-2007 Open Source Applications Foundation.
+/*
+ *   Copyright (c) 2007-2008 Open Source Applications Foundation
  *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions: 
+ *   Licensed under the Apache License, Version 2.0 (the "License");
+ *   you may not use this file except in compliance with the License.
+ *   You may obtain a copy of the License at
  *
- * The above copyright notice and this permission notice shall be included
- * in all copies or substantial portions of the Software. 
+ *       http://www.apache.org/licenses/LICENSE-2.0
  *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
- * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
- * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- * ====================================================================
+ *   Unless required by applicable law or agreed to in writing, software
+ *   distributed under the License is distributed on an "AS IS" BASIS,
+ *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ *   See the License for the specific language governing permissions and
+ *   limitations under the License.
  */
 
 #include <map>
+#include <string.h>
 #include <jni.h>
 
 #include "JCCEnv.h"
 
 extern JCCEnv *env;
 
-#ifdef PYTHON
-#include "JObject.h"
-#include "macros.h"
-#endif
-
-#ifdef _MSC_VER
-DWORD JCCEnv::VM_ENV = 0;
+#if defined(_MSC_VER) || defined(__WIN32)
+_DLL_EXPORT DWORD VM_ENV = 0;
 #else
 pthread_key_t JCCEnv::VM_ENV = (pthread_key_t) NULL;
 #endif
 
-#ifdef _MSC_VER
+#if defined(_MSC_VER) || defined(__WIN32)
 
 static CRITICAL_SECTION *mutex = NULL;
 
@@ -71,7 +60,7 @@ public:
 
 JCCEnv::JCCEnv(JavaVM *vm, JNIEnv *vm_env)
 {
-#ifdef _MSC_VER
+#if defined(_MSC_VER) || defined(__WIN32)
     if (!mutex)
     {
         mutex = new CRITICAL_SECTION();
@@ -91,7 +80,11 @@ JCCEnv::JCCEnv(JavaVM *vm, JNIEnv *vm_env)
 
     _sys = (jclass) vm_env->NewGlobalRef(vm_env->FindClass("java/lang/System"));
     _obj = (jclass) vm_env->NewGlobalRef(vm_env->FindClass("java/lang/Object"));
+#ifdef _jcc_lib
+    _thr = (jclass) vm_env->NewGlobalRef(vm_env->FindClass("org/osafoundation/jcc/PythonException"));
+#else
     _thr = (jclass) vm_env->NewGlobalRef(vm_env->FindClass("java/lang/RuntimeException"));
+#endif
     _mids = new jmethodID[max_mid];
 
     _mids[mid_sys_identityHashCode] =
@@ -109,12 +102,9 @@ JCCEnv::JCCEnv(JavaVM *vm, JNIEnv *vm_env)
     _mids[mid_obj_getClass] =
         vm_env->GetMethodID(_obj, "getClass",
                             "()Ljava/lang/Class;");
-    _mids[mid_thr_getMessage] =
-        vm_env->GetMethodID(_thr, "getMessage",
-                            "()Ljava/lang/String;");
 }
 
-#ifdef _MSC_VER
+#if defined(_MSC_VER) || defined(__WIN32)
 
 void JCCEnv::set_vm_env(JNIEnv *vm_env)
 {
@@ -155,29 +145,34 @@ jobject JCCEnv::newGlobalRef(jobject obj, int id)
 {
     if (obj)
     {
-        lock locked;
+        if (id)  /* zero when weak global ref is desired */
+        {
+            lock locked;
 
-        for (std::multimap<int, countedRef>::iterator iter = refs.find(id);
-             iter != refs.end();
-             iter++) {
-            if (iter->first != id)
-                break;
-            if (isSame(obj, iter->second.global))
-            {
-                iter->second.count += 1;
-                return iter->second.global;
+            for (std::multimap<int, countedRef>::iterator iter = refs.find(id);
+                 iter != refs.end();
+                 iter++) {
+                if (iter->first != id)
+                    break;
+                if (isSame(obj, iter->second.global))
+                {
+                    iter->second.count += 1;
+                    return iter->second.global;
+                }
             }
+
+            countedRef ref;
+            JNIEnv *vm_env = get_vm_env();
+
+            ref.global = vm_env->NewGlobalRef(obj);
+            ref.count = 1;
+            refs.insert(std::pair<const int, countedRef>(id, ref));
+            vm_env->DeleteLocalRef(obj);
+
+            return ref.global;
         }
-
-        countedRef ref;
-        JNIEnv *vm_env = get_vm_env();
-
-        ref.global = vm_env->NewGlobalRef(obj);
-        ref.count = 1;
-        refs.insert(std::pair<int, countedRef>(id, ref));
-        vm_env->DeleteLocalRef(obj);
-
-        return ref.global;
+        else
+            return get_vm_env()->NewWeakGlobalRef(obj);
     }
 
     return NULL;
@@ -187,27 +182,32 @@ jobject JCCEnv::deleteGlobalRef(jobject obj, int id)
 {
     if (obj)
     {
-        lock locked;
+        if (id)  /* zero when obj is weak global ref */
+        {
+            lock locked;
 
-        for (std::multimap<int, countedRef>::iterator iter = refs.find(id);
-             iter != refs.end();
-             iter++) {
-            if (iter->first != id)
-                break;
-            if (isSame(obj, iter->second.global))
-            {
-                if (iter->second.count == 1)
+            for (std::multimap<int, countedRef>::iterator iter = refs.find(id);
+                 iter != refs.end();
+                 iter++) {
+                if (iter->first != id)
+                    break;
+                if (isSame(obj, iter->second.global))
                 {
-                    get_vm_env()->DeleteGlobalRef(iter->second.global);
-                    refs.erase(iter);
+                    if (iter->second.count == 1)
+                    {
+                        get_vm_env()->DeleteGlobalRef(iter->second.global);
+                        refs.erase(iter);
+                    }
+                    else
+                        iter->second.count -= 1;
+                    return NULL;
                 }
-                else
-                    iter->second.count -= 1;
-                return NULL;
             }
-        }
 
-        printf("deleting non-existent ref: 0x%x\n", id);
+            printf("deleting non-existent ref: 0x%x\n", id);
+        }
+        else
+            get_vm_env()->DeleteWeakGlobalRef(obj);
     }
 
     return NULL;
@@ -260,6 +260,10 @@ int JCCEnv::getArrayLength(jarray array)
     return len;
 }
 
+jclass JCCEnv::getPythonExceptionClass()
+{
+    return _thr;
+}
 
 void JCCEnv::reportException()
 {
@@ -272,9 +276,10 @@ void JCCEnv::reportException()
         vm_env->ExceptionClear();
 
 #ifdef PYTHON
-        /* kludge: if RuntimeException is thrown assume a Python error */
-        jobject cls =
-            vm_env->CallObjectMethod(throwable, _mids[mid_obj_getClass]);
+        /* _thr is PythonException ifdef _jcc_lib (shared mode)
+         * if not shared mode, _thr is RuntimeException
+         */
+        jobject cls = (jobject) vm_env->GetObjectClass(throwable);
 
         if (vm_env->IsSameObject(cls, _thr))
             throw pythonError(throwable);
@@ -497,10 +502,31 @@ DEFINE_SET_FIELD(jshort, Short)
 void JCCEnv::setClassPath(const char *classPath)
 {
     JNIEnv *vm_env = get_vm_env();
-    jstring key = vm_env->NewStringUTF("java.class.path");
-    jstring cp = vm_env->NewStringUTF(classPath);
+    jclass _ucl = (jclass) vm_env->FindClass("java/net/URLClassLoader");
+    jclass _fil = (jclass) vm_env->FindClass("java/io/File");
+    jmethodID mid = vm_env->GetStaticMethodID(_ucl, "getSystemClassLoader",
+                                              "()Ljava/lang/ClassLoader;");
+    jobject classLoader = vm_env->CallStaticObjectMethod(_ucl, mid);
+    jmethodID mf = vm_env->GetMethodID(_fil, "<init>", "(Ljava/lang/String;)V");
+    jmethodID mu = vm_env->GetMethodID(_fil, "toURL", "()Ljava/net/URL;");
+    jmethodID ma = vm_env->GetMethodID(_ucl, "addURL", "(Ljava/net/URL;)V");
+#ifdef WINDOWS
+    char *pathsep = ";";
+#else
+    char *pathsep = ":";
+#endif
+    char *path = strdup(classPath);
 
-    vm_env->CallStaticObjectMethod(_sys, _mids[mid_sys_setProperty], key, cp);
+    for (char *cp = strtok(path, pathsep);
+         cp != NULL;
+         cp = strtok(NULL, pathsep)) {
+        jstring string = vm_env->NewStringUTF(cp);
+        jobject file = vm_env->NewObject(_fil, mf, string);
+        jobject url = vm_env->CallObjectMethod(file, mu);
+
+        vm_env->CallVoidMethod(classLoader, ma, url);
+    }
+    free(path);
 }
 
 jstring JCCEnv::fromUTF(const char *bytes)
@@ -538,6 +564,13 @@ char *JCCEnv::toString(jobject obj)
         : NULL;
 }
 
+char *JCCEnv::getClassName(jobject obj)
+{
+    return obj
+        ? toString(callObjectMethod(obj, _mids[mid_obj_getClass]))
+        : NULL;
+}
+
 #ifdef PYTHON
 
 jstring JCCEnv::fromPyString(PyObject *object)
@@ -619,61 +652,13 @@ PyObject *JCCEnv::fromJString(jstring js)
     }
 }
 
-extern PyObject *PyErr_SetJavaError(jthrowable throwable);
-
-jobjectArray JCCEnv::fromPySequence(jclass cls, PyObject *sequence)
+/* may be called from finalizer thread which has no vm_env thread local */
+void JCCEnv::finalizeObject(JNIEnv *jenv, PyObject *obj)
 {
-    if (sequence == Py_None)
-        return NULL;
+    PythonGIL gil;
 
-    if (!PySequence_Check(sequence))
-    {
-        PyErr_SetObject(PyExc_TypeError, sequence);
-        return NULL;
-    }
-
-    int length = PySequence_Length(sequence);
-    jobjectArray array;
-
-    try {
-        array = newObjectArray(cls, length);
-    } catch (JCCEnv::pythonError) {
-        return NULL;
-    } catch (JCCEnv::exception e) {
-        PyErr_SetJavaError(e.throwable);
-        return NULL;
-    }
-
-    for (int i = 0; i < length; i++) {
-        PyObject *obj = PySequence_GetItem(sequence, i);
-        jobject jobj;
-
-        if (!obj)
-            break;
-        else if (obj == Py_None)
-            jobj = NULL;
-        else if (PyString_Check(obj) || PyUnicode_Check(obj))
-            jobj = fromPyString(obj);
-        else if (PyObject_TypeCheck(obj, &JObjectType))
-            jobj = ((t_JObject *) obj)->object.this$;
-        else
-        {
-            PyErr_SetObject(PyExc_TypeError, obj);
-            Py_DECREF(obj);
-            return NULL;
-        }
-
-        Py_DECREF(obj);
-
-        try {
-            setObjectArrayElement(array, i, jobj);
-        } catch (JCCEnv::exception e) {
-            PyErr_SetJavaError(e.throwable);
-            return NULL;
-        }
-    }
-
-    return array;
+    set_vm_env(jenv);
+    Py_DECREF(obj);
 }
 
-#endif
+#endif /* PYTHON */
