@@ -1,24 +1,17 @@
-/* ====================================================================
- * Copyright (c) 2007-2007 Open Source Applications Foundation.
+/*
+ *   Copyright (c) 2007-2008 Open Source Applications Foundation
  *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions: 
+ *   Licensed under the Apache License, Version 2.0 (the "License");
+ *   you may not use this file except in compliance with the License.
+ *   You may obtain a copy of the License at
  *
- * The above copyright notice and this permission notice shall be included
- * in all copies or substantial portions of the Software. 
+ *       http://www.apache.org/licenses/LICENSE-2.0
  *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
- * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
- * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- * ====================================================================
+ *   Unless required by applicable law or agreed to in writing, software
+ *   distributed under the License is distributed on an "AS IS" BASIS,
+ *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ *   See the License for the specific language governing permissions and
+ *   limitations under the License.
  */
 
 #include <jni.h>
@@ -28,7 +21,10 @@
 #include "java/lang/Class.h"
 #include "java/lang/String.h"
 #include "java/lang/Throwable.h"
-#include "java/lang/RuntimeException.h"
+#include "java/lang/Boolean.h"
+#include "java/lang/Integer.h"
+#include "java/lang/Long.h"
+#include "java/lang/Double.h"
 #include "java/util/Iterator.h"
 #include "JArray.h"
 #include "functions.h"
@@ -40,7 +36,37 @@ using namespace java::util;
 PyObject *PyExc_JavaError = PyExc_ValueError;
 PyObject *PyExc_InvalidArgsError = PyExc_ValueError;
 
-#ifdef _MSC_VER
+PyObject *_setExceptionTypes(PyObject *self, PyObject *args)
+{
+    if (!PyArg_ParseTuple(args, "OO",
+                          &PyExc_JavaError, &PyExc_InvalidArgsError))
+        return NULL;
+
+    Py_RETURN_NONE;
+}
+
+PyObject *findClass(PyObject *self, PyObject *args)
+{
+    char *className;
+
+    if (!PyArg_ParseTuple(args, "s", &className))
+        return NULL;
+
+    try {
+        jclass cls = env->findClass(className);
+
+        if (cls)
+            return t_Class::wrap_Object(Class(cls));
+    } catch (JCCEnv::exception e) {
+        PyErr_SetJavaError(e.throwable);
+        return NULL;
+    }
+
+    Py_RETURN_NONE;
+}
+
+
+#if defined(_MSC_VER) || defined(__SUNPRO_CC)
 int __parseArgs(PyObject *args, char *types, ...)
 {
     int count = ((PyTupleObject *)(args))->ob_size;
@@ -107,8 +133,17 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
               if (arg == Py_None)
                   break;
 
+              /* ensure that class Class is initialized (which may not be the
+               * case because of earlier recursion avoidance (JObject(cls)).
+               */
+              if (!Class::class$)
+                  Class::initializeClass();
+
               if (array)
               {
+                  if (PyObject_TypeCheck(arg, JArrayObjectType))
+                      break;
+
                   if (PySequence_Check(arg) &&
                       !PyString_Check(arg) && !PyUnicode_Check(arg))
                   {
@@ -131,14 +166,45 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
               else if (PyObject_TypeCheck(arg, &ObjectType) &&
                        cls->isInstance(((t_Object *) arg)->object))
                   break;
+              else if (PyObject_TypeCheck(arg, &FinalizerProxyType))
+              {
+                  arg = ((t_fp *) arg)->object;
+                  if (PyObject_TypeCheck(arg, &ObjectType) &&
+                      cls->isInstance(((t_Object *) arg)->object))
+                      break;
+              }
 
               return -1;
           }
 
           case 'Z':           /* boolean, strict */
           {
-              if (arg == Py_True || arg == Py_False)
+              if (array)
+              {
+                  if (arg == Py_None)
+                      break;
+
+                  if (PyObject_TypeCheck(arg, JArrayBoolType))
+                      break;
+
+                  if (PySequence_Check(arg))
+                  {
+                      if (PySequence_Length(arg) > 0)
+                      {
+                          PyObject *obj = PySequence_GetItem(arg, 0);
+                          int ok = obj == Py_True || obj == Py_False;
+
+                          Py_DECREF(obj);
+                          if (ok)
+                              break;
+                      }
+                      else
+                          break;
+                  }
+              }
+              else if (arg == Py_True || arg == Py_False)
                   break;
+
               return -1;
           }
 
@@ -146,6 +212,10 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
           {
               if (array)
               {
+                  if (arg == Py_None)
+                      break;
+                  if (PyObject_TypeCheck(arg, JArrayByteType))
+                      break;
                   if (arg == Py_None || PyString_Check(arg))
                       break;
               }
@@ -158,8 +228,11 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
           {
               if (array)
               {
-                  if (arg == Py_None || PyString_Check(arg) ||
-                      PyUnicode_Check(arg))
+                  if (arg == Py_None)
+                      break;
+                  if (PyObject_TypeCheck(arg, JArrayCharType))
+                      break;
+                  if (arg == Py_None || PyUnicode_Check(arg))
                       break;
               }
               else if (PyUnicode_Check(arg) && PyUnicode_GET_SIZE(arg) == 1)
@@ -168,10 +241,15 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
           }
 
           case 'I':           /* int */
-          case 'S':           /* short */
           {
               if (array)
               {
+                  if (arg == Py_None)
+                      break;
+
+                  if (PyObject_TypeCheck(arg, JArrayIntType))
+                      break;
+
                   if (PySequence_Check(arg))
                   {
                       if (PySequence_Length(arg) > 0)
@@ -189,14 +267,51 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
               }
               else if (PyInt_CheckExact(arg))
                   break;
+
+              return -1;
+          }
+
+          case 'S':           /* short */
+          {
+              if (array)
+              {
+                  if (arg == Py_None)
+                      break;
+
+                  if (PyObject_TypeCheck(arg, JArrayShortType))
+                      break;
+
+                  if (PySequence_Check(arg))
+                  {
+                      if (PySequence_Length(arg) > 0)
+                      {
+                          PyObject *obj = PySequence_GetItem(arg, 0);
+                          int ok = PyInt_CheckExact(obj);
+
+                          Py_DECREF(obj);
+                          if (ok)
+                              break;
+                      }
+                      else
+                          break;
+                  }
+              }
+              else if (PyInt_CheckExact(arg))
+                  break;
+
               return -1;
           }
 
           case 'D':           /* double */
-          case 'F':           /* float */
           {
               if (array)
               {
+                  if (arg == Py_None)
+                      break;
+
+                  if (PyObject_TypeCheck(arg, JArrayDoubleType))
+                      break;
+
                   if (PySequence_Check(arg))
                   {
                       if (PySequence_Length(arg) > 0)
@@ -214,6 +329,38 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
               }
               else if (PyFloat_CheckExact(arg))
                   break;
+
+              return -1;
+          }
+
+          case 'F':           /* float */
+          {
+              if (array)
+              {
+                  if (arg == Py_None)
+                      break;
+
+                  if (PyObject_TypeCheck(arg, JArrayFloatType))
+                      break;
+
+                  if (PySequence_Check(arg))
+                  {
+                      if (PySequence_Length(arg) > 0)
+                      {
+                          PyObject *obj = PySequence_GetItem(arg, 0);
+                          int ok = PyFloat_CheckExact(obj);
+
+                          Py_DECREF(obj);
+                          if (ok)
+                              break;
+                      }
+                      else
+                          break;
+                  }
+              }
+              else if (PyFloat_CheckExact(arg))
+                  break;
+
               return -1;
           }
 
@@ -221,6 +368,12 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
           {
               if (array)
               {
+                  if (arg == Py_None)
+                      break;
+
+                  if (PyObject_TypeCheck(arg, JArrayLongType))
+                      break;
+
                   if (PySequence_Check(arg))
                   {
                       if (PySequence_Length(arg) > 0)
@@ -238,6 +391,7 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
               }
               else if (PyLong_CheckExact(arg))
                   break;
+
               return -1;
           }
 
@@ -247,6 +401,10 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
               {
                   if (arg == Py_None)
                       break;
+
+                  if (PyObject_TypeCheck(arg, JArrayStringType))
+                      break;
+
                   if (PySequence_Check(arg) && 
                       !PyString_Check(arg) && !PyUnicode_Check(arg))
                   {
@@ -268,9 +426,13 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
               else if (arg == Py_None ||
                        PyString_Check(arg) || PyUnicode_Check(arg))
                   break;
+
               return -1;
           }
 
+          case 'o':         /* java.lang.Object */
+            break;
+
           default:
             return -1;
         }
@@ -297,18 +459,20 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
               break;
           }
 
-          case 'j':           /* Java object */
+          case 'j':           /* Java object except String and Object */
           {
               Class *cls = va_arg(check, Class *);
 
               if (array)
               {
-                  JArray<Object> *array = va_arg(list, JArray<Object> *);
+                  JArray<jobject> *array = va_arg(list, JArray<jobject> *);
 
                   if (arg == Py_None)
-                      *array = JArray<Object>((jobject) NULL);
+                      *array = JArray<jobject>((jobject) NULL);
+                  else if (PyObject_TypeCheck(arg, JArrayObjectType))
+                      *array = ((t_jarray<jobject> *) arg)->array;
                   else 
-                      *array = JArray<Object>((jclass) cls->this$, arg);
+                      *array = JArray<jobject>((jclass) cls->this$, arg);
 
                   if (PyErr_Occurred())
                       return -1;
@@ -317,6 +481,9 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
               {
                   Object *obj = va_arg(list, Object *);
 
+                  if (PyObject_TypeCheck(arg, &FinalizerProxyType))
+                      arg = ((t_fp *) arg)->object;
+
                   *obj = arg == Py_None
                       ? Object(NULL)
                       : ((t_Object *) arg)->object;
@@ -324,6 +491,30 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
               break;
           }
 
+          case 'Z':           /* boolean, strict */
+          {
+              if (array)
+              {
+                  JArray<jboolean> *array = va_arg(list, JArray<jboolean> *);
+
+                  if (arg == Py_None)
+                      *array = JArray<jboolean>((jobject) NULL);
+                  else if (PyObject_TypeCheck(arg, JArrayBoolType))
+                      *array = ((t_jarray<jboolean> *) arg)->array;
+                  else
+                      *array = JArray<jboolean>(arg);
+
+                  if (PyErr_Occurred())
+                      return -1;
+              }
+              else
+              {
+                  jboolean *b = va_arg(list, jboolean *);
+                  *b = arg == Py_True;
+              }
+              break;
+          }
+
           case 'B':           /* byte */
           {
               if (array)
@@ -332,6 +523,8 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
 
                   if (arg == Py_None)
                       *array = JArray<jbyte>((jobject) NULL);
+                  else if (PyObject_TypeCheck(arg, JArrayByteType))
+                      *array = ((t_jarray<jbyte> *) arg)->array;
                   else 
                       *array = JArray<jbyte>(arg);
 
@@ -346,28 +539,6 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
               break;
           }
 
-          case 'Z':           /* boolean, strict */
-          {
-              if (array)
-              {
-                  JArray<jboolean> *array = va_arg(list, JArray<jboolean> *);
-
-                  if (arg == Py_None)
-                      *array = JArray<jboolean>((jobject) NULL);
-                  else 
-                      *array = JArray<jboolean>(arg);
-
-                  if (PyErr_Occurred())
-                      return -1;
-              }
-              else
-              {
-                  jboolean *b = va_arg(list, jboolean *);
-                  *b = arg == Py_True;
-              }
-              break;
-          }
-
           case 'C':           /* char */
           {
               if (array)
@@ -376,6 +547,8 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
 
                   if (arg == Py_None)
                       *array = JArray<jchar>((jobject) NULL);
+                  else if (PyObject_TypeCheck(arg, JArrayCharType))
+                      *array = ((t_jarray<jchar> *) arg)->array;
                   else 
                       *array = JArray<jchar>(arg);
 
@@ -398,6 +571,8 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
 
                   if (arg == Py_None)
                       *array = JArray<jint>((jobject) NULL);
+                  else if (PyObject_TypeCheck(arg, JArrayIntType))
+                      *array = ((t_jarray<jint> *) arg)->array;
                   else 
                       *array = JArray<jint>(arg);
 
@@ -420,6 +595,8 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
 
                   if (arg == Py_None)
                       *array = JArray<jshort>((jobject) NULL);
+                  else if (PyObject_TypeCheck(arg, JArrayShortType))
+                      *array = ((t_jarray<jshort> *) arg)->array;
                   else 
                       *array = JArray<jshort>(arg);
 
@@ -436,22 +613,73 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
 
           case 'D':           /* double */
           {
-              jdouble *d = va_arg(list, jdouble *);
-              *d = (jdouble) PyFloat_AsDouble(arg);
+              if (array)
+              {
+                  JArray<jdouble> *array = va_arg(list, JArray<jdouble> *);
+
+                  if (arg == Py_None)
+                      *array = JArray<jdouble>((jobject) NULL);
+                  else if (PyObject_TypeCheck(arg, JArrayDoubleType))
+                      *array = ((t_jarray<jdouble> *) arg)->array;
+                  else 
+                      *array = JArray<jdouble>(arg);
+
+                  if (PyErr_Occurred())
+                      return -1;
+              }
+              else
+              {
+                  jdouble *d = va_arg(list, jdouble *);
+                  *d = (jdouble) PyFloat_AsDouble(arg);
+              }
               break;
           }
 
           case 'F':           /* float */
           {
-              jfloat *d = va_arg(list, jfloat *);
-              *d = (jfloat) (float) PyFloat_AsDouble(arg);
+              if (array)
+              {
+                  JArray<jfloat> *array = va_arg(list, JArray<jfloat> *);
+
+                  if (arg == Py_None)
+                      *array = JArray<jfloat>((jobject) NULL);
+                  else if (PyObject_TypeCheck(arg, JArrayFloatType))
+                      *array = ((t_jarray<jfloat> *) arg)->array;
+                  else 
+                      *array = JArray<jfloat>(arg);
+
+                  if (PyErr_Occurred())
+                      return -1;
+              }
+              else
+              {
+                  jfloat *d = va_arg(list, jfloat *);
+                  *d = (jfloat) (float) PyFloat_AsDouble(arg);
+              }
               break;
           }
 
           case 'J':           /* long long */
           {
-              jlong *l = va_arg(list, jlong *);
-              *l = (jlong) PyLong_AsLongLong(arg);
+              if (array)
+              {
+                  JArray<jlong> *array = va_arg(list, JArray<jlong> *);
+
+                  if (arg == Py_None)
+                      *array = JArray<jlong>((jobject) NULL);
+                  else if (PyObject_TypeCheck(arg, JArrayLongType))
+                      *array = ((t_jarray<jlong> *) arg)->array;
+                  else 
+                      *array = JArray<jlong>(arg);
+
+                  if (PyErr_Occurred())
+                      return -1;
+              }
+              else
+              {
+                  jlong *l = va_arg(list, jlong *);
+                  *l = (jlong) PyLong_AsLongLong(arg);
+              }
               break;
           }
 
@@ -459,12 +687,14 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
           {
               if (array)
               {
-                  JArray<String> *array = va_arg(list, JArray<String> *);
+                  JArray<jstring> *array = va_arg(list, JArray<jstring> *);
 
                   if (arg == Py_None)
-                      *array = JArray<String>((jobject) NULL);
+                      *array = JArray<jstring>((jobject) NULL);
+                  else if (PyObject_TypeCheck(arg, JArrayStringType))
+                      *array = ((t_jarray<jstring> *) arg)->array;
                   else 
-                      *array = JArray<String>(arg);
+                      *array = JArray<jstring>(arg);
 
                   if (PyErr_Occurred())
                       return -1;
@@ -485,6 +715,66 @@ int _parseArgs(PyObject **args, unsigned int count, char *types, ...)
               break;
           }
 
+          case 'o':           /* java.lang.Object  */
+          {
+              if (array)
+              {
+                  JArray<Object> *array = va_arg(list, JArray<Object> *);
+
+                  if (arg == Py_None)
+                      *array = JArray<Object>((jobject) NULL);
+                  else 
+                      *array = JArray<Object>(arg);
+
+                  if (PyErr_Occurred())
+                      return -1;
+              }
+              else
+              {
+                  Object *obj = va_arg(list, Object *);
+
+                  if (arg == Py_None)
+                      *obj = Object(NULL);
+                  else if (PyObject_TypeCheck(arg, &ObjectType))
+                      *obj = ((t_Object *) arg)->object;
+                  else if (PyObject_TypeCheck(arg, &FinalizerProxyType))
+                  {
+                      arg = ((t_fp *) arg)->object;
+                      if (PyObject_TypeCheck(arg, &ObjectType))
+                          *obj = ((t_Object *) arg)->object;
+                      else
+                          return -1;
+                  }
+                  else if (PyString_Check(arg) || PyUnicode_Check(arg))
+                  {
+                      *obj = p2j(arg);
+                      if (PyErr_Occurred())
+                          return -1;
+                  }
+                  else if (arg == Py_True)
+                      *obj = *Boolean::TRUE;
+                  else if (arg == Py_False)
+                      *obj = *Boolean::FALSE;
+                  else if (PyInt_Check(arg))
+                  {
+                      long ln = PyInt_AS_LONG(arg);
+                      int n = (int) ln;
+
+                      if (ln != (long) n)
+                          *obj = Long((jlong) ln);
+                      else
+                          *obj = Integer((jint) n);
+                  }
+                  else if (PyLong_Check(arg))
+                      *obj = Long((jlong) PyLong_AsLongLong(arg));
+                  else if (PyFloat_Check(arg))
+                      *obj = Double((jdouble) PyFloat_AS_DOUBLE(arg));
+                  else
+                      return -1;
+              }
+              break;
+          }
+
           default:
             return -1;
         }
@@ -543,7 +833,7 @@ PyObject *PyErr_SetArgsError(PyTypeObject *type, char *name, PyObject *args)
 
 PyObject *PyErr_SetJavaError(jthrowable throwable)
 {
-    PyObject *err = t_Throwable::wrapObject(Throwable(throwable));
+    PyObject *err = t_Throwable::wrap_Object(Throwable(throwable));
 
     PyErr_SetObject(PyExc_JavaError, err);
     Py_DECREF(err);
@@ -585,17 +875,28 @@ void throwPythonError(void)
             }
         }
         else
+        {
+            Py_DECREF(exc);
             Py_XDECREF(traceback);
+        }
     }
-
-    if (exc && PyErr_GivenExceptionMatches(exc, PyExc_StopIteration))
+    else if (exc && PyErr_GivenExceptionMatches(exc, PyExc_StopIteration))
     {
         PyErr_Clear();
         return;
     }
 
-    env->get_vm_env()->ThrowNew(RuntimeException::initializeClass(),
-                                "PythonError");
+    if (exc)
+    {
+        PyObject *name = PyObject_GetAttrString(exc, "__name__");
+
+        env->get_vm_env()->ThrowNew(env->getPythonExceptionClass(),
+                                    PyString_AS_STRING(name));
+        Py_DECREF(name);
+    }
+    else
+        env->get_vm_env()->ThrowNew(env->getPythonExceptionClass(),
+                                    "python error");
 }
 
 void throwTypeError(const char *name, PyObject *object)
@@ -605,8 +906,7 @@ void throwTypeError(const char *name, PyObject *object)
     PyErr_SetObject(PyExc_TypeError, tuple);
     Py_DECREF(tuple);
 
-    env->get_vm_env()->ThrowNew(RuntimeException::initializeClass(),
-                                "PythonError");
+    env->get_vm_env()->ThrowNew(env->getPythonExceptionClass(), "type error");
 }
 
 int abstract_init(PyObject *self, PyObject *args, PyObject *kwds)
@@ -623,8 +923,9 @@ int abstract_init(PyObject *self, PyObject *args, PyObject *kwds)
 PyObject *callSuper(PyTypeObject *type, const char *name, PyObject *args,
                     int cardinality)
 {
-    PyObject *method = PyObject_GetAttrString((PyObject *) type->tp_base,
-                                              (char *) name); // python 2.4
+    PyObject *super = (PyObject *) type->tp_base;
+    PyObject *method =
+        PyObject_GetAttrString(super, (char *) name); // python 2.4 cast
     PyObject *value;
 
     if (!method)
@@ -663,7 +964,8 @@ PyObject *callSuper(PyTypeObject *type, PyObject *self,
     if (!super)
         return NULL;
 
-    method = PyObject_GetAttrString(super, (char *) name); // python 2.4
+    method = PyObject_GetAttrString(super, (char *) name); // python 2.4 cast
+    Py_DECREF(super);
     if (!method)
         return NULL;
 
@@ -685,13 +987,16 @@ PyObject *callSuper(PyTypeObject *type, PyObject *self,
     return value;
 }
 
-int castCheck(PyObject *obj, jclass cls, int reportError)
+PyObject *castCheck(PyObject *obj, jclass cls, int reportError)
 {
+    if (PyObject_TypeCheck(obj, &FinalizerProxyType))
+        obj = ((t_fp *) obj)->object;
+
     if (!PyObject_TypeCheck(obj, &ObjectType))
     {
         if (reportError)
             PyErr_SetObject(PyExc_TypeError, obj);
-        return -1;
+        return NULL;
     }
 
     jobject jobj = ((t_Object *) obj)->object.this$;
@@ -700,9 +1005,88 @@ int castCheck(PyObject *obj, jclass cls, int reportError)
     {
         if (reportError)
             PyErr_SetObject(PyExc_TypeError, obj);
-        return -1;
+        return NULL;
     }
 
-    return 0;
+    return obj;
 }
 
+PyObject *get_extension_iterator(PyObject *self)
+{
+    return PyObject_CallMethod(self, "iterator", "");
+}
+
+PyObject *get_extension_next(PyObject *self)
+{
+    return PyObject_CallMethod(self, "next", "");
+}
+
+PyObject *get_extension_nextElement(PyObject *self)
+{
+    return PyObject_CallMethod(self, "nextElement", "");
+}
+
+jobjectArray fromPySequence(jclass cls, PyObject *sequence)
+{
+    if (sequence == Py_None)
+        return NULL;
+
+    if (!PySequence_Check(sequence))
+    {
+        PyErr_SetObject(PyExc_TypeError, sequence);
+        return NULL;
+    }
+
+    int length = PySequence_Length(sequence);
+    jobjectArray array;
+
+    try {
+        array = env->newObjectArray(cls, length);
+    } catch (JCCEnv::pythonError) {
+        return NULL;
+    } catch (JCCEnv::exception e) {
+        PyErr_SetJavaError(e.throwable);
+        return NULL;
+    }
+
+    JNIEnv *vm_env = env->get_vm_env();
+
+    for (int i = 0; i < length; i++) {
+        PyObject *obj = PySequence_GetItem(sequence, i);
+        int fromString = 0;
+        jobject jobj;
+
+        if (!obj)
+            break;
+        else if (obj == Py_None)
+            jobj = NULL;
+        else if (PyString_Check(obj) || PyUnicode_Check(obj))
+        {
+            jobj = env->fromPyString(obj);
+            fromString = 1;
+        }
+        else if (PyObject_TypeCheck(obj, &JObjectType))
+            jobj = ((t_JObject *) obj)->object.this$;
+        else if (PyObject_TypeCheck(obj, &FinalizerProxyType))
+            jobj = ((t_JObject *) ((t_fp *) obj)->object)->object.this$;
+        else /* todo: add auto-boxing of primitive types */
+        {
+            PyErr_SetObject(PyExc_TypeError, obj);
+            Py_DECREF(obj);
+            return NULL;
+        }
+
+        Py_DECREF(obj);
+
+        try {
+            env->setObjectArrayElement(array, i, jobj);
+            if (fromString)
+                vm_env->DeleteLocalRef(jobj);
+        } catch (JCCEnv::exception e) {
+            PyErr_SetJavaError(e.throwable);
+            return NULL;
+        }
+    }
+
+    return array;
+}
